
PARTITION_STYLE_MBR = 0
PARTITION_STYLE_GPT = 1
PARTITION_STYLE_RAW = 2

PARTITION_ENTRY_UNUSED = 0x00
PARTITION_FAT_12 = 0x01
PARTITION_XENIX_1 = 0x02
PARTITION_XENIX_2 = 0x03
PARTITION_FAT_16 = 0x04
PARTITION_EXTENDED = 0x05
PARTITION_HUGE = 0x06
PARTITION_IFS = 0x07
PARTITION_OS2BOOTMGR = 0x0A
PARTITION_FAT32 = 0x0B
PARTITION_FAT32_XINT13 = 0x0C
PARTITION_XINT13 = 0x0E
PARTITION_XINT13_EXTENDED = 0x0F
PARTITION_PREP = 0x41
PARTITION_LDM = 0x42
PARTITION_UNIX = 0x63

TRUE = 1

DISK_ATTRIBUTE_OFFLINE = 0x0000000000000001
DISK_ATTRIBUTE_READ_ONLY = 0x0000000000000002

DISK_SAN_POLICY_UNKNOWN = 0
DISK_SAN_POLICY_ONLINE = 1
DISK_SAN_POLICY_OFFLINE_SHARED = 2
DISK_SAN_POLICY_OFFLINE = 3
DISK_SAN_POLICY_MAX = 4

MOUNTMGR_AUTO_MOUNT_STATE_DISABLED = 0
MOUNTMGR_AUTO_MOUNT_STATE_ENABLED = 1

