__version__ = "0.3.9"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.3.9)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '69e422900e88db0c1ce10ee38ccd03b5bb34338f'
__git_head_subject__ = 'Finished release v0.3.9.'
__git_head_message__ = ''
__git_dirty_diff__ = ''